"use strict";

var superagent = require('superagent');
var Immutable = require('immutable');
var _ = require('underscore');

var LegacyI3APIHeaders = require('legacy/I3APIHeaders');
var Router = require('Router');
var Wish = require('Wish');
var Logger = require('Logger');

// Action factories

function Retrieve(path, method, url, mapper, options) {
    mapper = mapper || ((x) => x);
    return {
        name: "API/Retrieve",
        transaction: (ws) => ws.setIn(path, Wish.Pending()),
        effect: (world) => API.request(world, url,
            _.partial(RetrievalWasResolved, path, mapper),
            _.partial(RetrievalWasRejected, path),
            options)
    };
}

function RetrievalWasResolved(path, mapper, value) {
    return {
        name: 'API/RetrievalWasResolved',
        transaction: (ws) => ws.setIn(path, Wish.Available(mapper(value)))
    };
}

function RetrievalWasRejected(path, error) {
    return {
        name: 'API/RetrievalWasRejected',
        transaction: (ws) => reportFatalError(ws, path, error)
    };
}

function reportFatalError(ws, path, error) {
    if (error.get('exception')) {
        Logger.errorLogger.captureException(error.get('exception'), {
            extra: {
                description: 'API Request Exception',
                path
            }
        });
    } else {
        var headers = error.get('headers').toJS();
        var request = error.get('req').toJS();

        Logger.errorLogger.captureMessage(
            'Fatal XRSP client error.',
            {
                extra: { headers, request }
            }
        );
    }
    return ws.setIn(path, Wish.Error(error))
        .update('FatalErrors', fatalErrors => fatalErrors.push(path))
        .set('LastSentryId', Wish.Pending());
}

// Helper functions

function errorFromSuperagentErrorAndRes(error, res) {
    // TODO:eo come up with a generalized approach to creating
    // error values when UI/application needs arise
    if (error) {
        return {
            exception: error,
            message: error.message,
            line: error.line,
            stack: error.stack,
            name: error.name,
        };
    } else if (res && !res.ok) {
        return {
            req: {
                url: res.req.url,
                headers: res.req.header
            },
            headers: res.header,
            status: res.status,
            body: res.body
        };
    }
}

function forwardedHeaders(world) {
    var headers = world.getIn(['ServerOnly', 'Request', 'headers']);
    return headers ? headers.toJS() : {};
}

// The main event

var API = {
    createErrorHandler: function(actionName, effect = function() {}) {
        return function (path, error) {
            return {
                name: actionName,
                transaction: (ws) => {
                    if (error.get('exception')) { // that's an error from JS, not the API
                        throw error.get('exception');
                    } else {
                        return ws.setIn(
                            // 4xx has meaning to the application, 5xx is just broken
                            path, error.get('status').toString().match(/^4/)
                                ? Wish.Available(error.set('error', true))
                                : Wish.Error(error));
                    }
                },
                effect: (wld) => effect(wld, error)
            };
        };
    },

    request: function(world, route, resolvingActionCreator, rejectingActionCreator, options) {
        options = options || {};
        var accept = options.accept || 'application/json';
        var method = options.method || 'GET';
        var payload = options.payload;
        var type = options.type || 'json';
        var extraHeaders = options.extraHeaders || {};

        var url = Router.urlForRoute(route);
        Logger.log('API.%s: %s', method, url);
        (new superagent.Request(method, url))
        .set(forwardedHeaders(world))
        .set('Accept', accept)
        .set(extraHeaders)
        .set('X-Requested-With', 'XMLHttpRequest')
        .type(type)
        .send(payload)
        .end(function(error, res) {
            error = errorFromSuperagentErrorAndRes(error, res);
            if (error) {
                Logger.error('API.%s ERROR: %s', method, JSON.stringify(error));
                world.performAction(rejectingActionCreator(Immutable.fromJS(error)));
            } else {
                world.performAction(resolvingActionCreator(
                    Immutable.fromJS(res.body),
                    Immutable.fromJS(res.headers)
                ));
            }
        });
    },

    get: function(world, path, route, mapper) {
        return Wish.performOnce(world, path, Retrieve(path, 'GET', route, mapper));
    },

    legacyGet: function(world, path, route, mapper) {
        return Wish.performOnce(world, path, Retrieve(path, 'GET', route, mapper, {extraHeaders: LegacyI3APIHeaders}));
    },

    // TODO: extract to a Constants dealer
    constants: function(world, ...constants) {
        return API.multiplex(world, 'api/v3/constants', 'Constants', constants);
    },

    multiplex: function(world, endpoint, rootPath, params) {
        var neededParams = {};
        var error = 0;
        var available = 0;

        params.forEach(param => {
            // XXX:cab Maybe check that current Wish didn't fail?
            if (!world.getIn([rootPath, param])) {
                world.performAction({
                    name: `API/MultiplexWillFetch[${rootPath}]?${param}`,
                    transaction: ws => ws.setIn(
                        [rootPath, param],
                        Wish.Pending()
                    )
                });
                neededParams[param] = true;
            } else if (world.getIn([rootPath, param]).get('state') === 'Available') {
                available++;
            } else if (world.getIn([rootPath, param]).get('state') === 'Error') {
                error++;
            }
        });

        if (error) {
            return Wish.Error();
        } else if (available === params.length) {
            return Wish.Available();
        } else if (Object.keys(neededParams).length === 0) {
            return Wish.Pending();
        }

        return API.request(
            world,
            Router.route(endpoint, neededParams),
            function resolve(body, headers) {
                return {
                    name: `API/MultiplexResolved[${rootPath}]`,
                    transaction: ws =>
                        Object.keys(neededParams).reduce((ws, param) => {
                            return ws.setIn(
                                [rootPath, param],
                                Wish.Available(body.getIn(['data', param]))
                            );
                        }, ws)
                };
            },
            function reject(error) {
                return {
                    name: `API/MultiplexRejected[${rootPath}]`,
                    transaction: ws =>
                        Object.keys(neededParams).reduce((ws, param) => {
                            var path = [rootPath, param];
                            return reportFatalError(ws, path, error);
                        }, ws)
                };
            }
        );
    }
};

module.exports = API;
