"use strict";

var _ = require('underscore');
var React = require('react/addons');
var Modal = require('react-modal');
var ReactIntl = require('react-intl');

var Client = require('dealers/Client');
var LoginSignup = require('dealers/LoginSignup');
var NavMenu = require('dealers/NavMenu');

var Analytics = require('elements/Analytics');
var Centered = require('elements/Centered');
var HtmlHead = require('elements/HtmlHead');
var Navigation = require('Navigation');
var PageFooter = require('elements/PageWrapper/PageFooter');
var PageWrapper = require('elements/PageWrapper/PageWrapper');
var ResizeWatcher = require('elements/ResizeWatcher.jsx');
var {UnexpectedError} = require('elements/Errors');
var Link = require('elements/Link');
var Look = require('elements/Look');

var {I18N} = require('I18N');
var Router = require('Router');
var SignupFlow = require('elements/LoginSignup/SignupFlow');
var Logger = require('Logger');
var World = require('World');


var ClickWrapper = React.createClass({
    mixins: [World.Child, Navigation],

    propTypes: {
        additionalClickHandler: React.PropTypes.func
    },

    getDefaultProps() {
        return {
            additionalClickHandler: () => {}
        };
    },

    onClick(event) {
        // Any links on the page which may be suitable for client-side
        // navigation, but which were not created with a <Link> will
        // be intercepted by navigateWithClick
        this.navigateWithClick(event);
        this.props.additionalClickHandler.call(this, event);
    },

    route: function(event) {
        if (event.target.tagName == 'A') {
            // TODO: query params?
            return Router.routeForUrlPath(event.target.pathname);
        }
    },

    render() {
        return <div onClick={this.onClick}>
            {this.props.children}
        </div>;
    }
});


const ModalWrapper = React.createClass({
    mixins: [World.Root],

    render() {
        return <div>{this.props.children}</div>;
    }
});


var Layout = React.createClass({
    mixins: [ReactIntl.IntlMixin, World.Root],

    componentDidMount: function () {
        this.world().onActionPerformed(() => {
            if (!this.measured) {
                Logger.stats.time('global.render', (cb) => this.forceUpdate(cb), ()=>{});
                this.measured = true;
            } else {
                this.forceUpdate();
            }
        });
        this.updateBodyClasses();
    },

    componentWillUpdate() {
        this.updateBodyClasses();
    },

    // Legitimate use of IDO_i5_world_do_not_use_this 1/2? :
    // setting the context for everyone else to catch via Worldly
    childContextTypes: {
        locales: React.PropTypes.arrayOf(React.PropTypes.string)
    },

    getChildContext() {
        return {
            locales: [this.props.world.get('Locale')]
        };
    },

    hideMobileNavMenuIfShown(event) {
        // don't intercept clicks inside the menu
        var mobileNavMenu = document.getElementById('mobile-user-menu');
        if (mobileNavMenu && mobileNavMenu.contains(event.target)) {
            return;
        }
        if (NavMenu.isMobileNavMenuOpen(this.world())) {
            event.preventDefault();
            NavMenu.toggleMobileNavMenu(this.world());
        }
    },

    render: function () {
        var fatalErrors = this.world().worldState.get('FatalErrors');

        return <ClickWrapper additionalClickHandler={this.hideMobileNavMenuIfShown}>
            <ResizeWatcher onResize={this.onResize} />
            <Modal isOpen={!!fatalErrors.size || LoginSignup.isModalSignupOpen(this.world())}
                   closeTimeoutMS={300}
                   onRequestClose={(event) => event.key === 'Escape' && this.closeSignup() }
                   contentWillFocus={this.onModalContentWillFocus}>
                <ModalWrapper world={this.world()}>
                    {/* TODO: ClickWrapper links should close the Modal! */}
                    <ClickWrapper>
                        <Centered>
                        {
                            !!fatalErrors.size ?
                            <UnexpectedError>
                                <Look.Quick wish={this.world().get('LastSentryId')}>{
                                    sentryId =>
                                        <I18N>Please <Link endpoint="contact-us" values={{sentryId}}>
                                            contact us
                                        </Link> or try again later.</I18N>
                                }</Look.Quick>
                            </UnexpectedError> :
                            [
                                <SignupFlow ref="flow" onClose={this.closeSignup} />,
                                <a className="reactmodal-close" title="close" href="#" onClick={this.closeSignup}></a>
                            ]
                        }
                        </Centered>
                        <a className="reactmodal-close" title="close" href="#" onClick={this.closeSignup}></a>
                    </ClickWrapper>
                </ModalWrapper>
            </Modal>
            <PageWrapper>
                <HtmlHead info={this.pageComponentType().htmlHead(this.world())} />
                {this.pageComponent()}
            </PageWrapper>
            <PageFooter />
            <Analytics />
        </ClickWrapper>;
    },

    onModalContentWillFocus: function() {
        if (document.addEventListener) {
            document.addEventListener('keydown', this.closeSignupListener);
        } else {
            window.attachEvent('onkeydown', this.closeSignupListener);
        }
        // focus first field in flow
        var flow = this.refs.flow;
        if (flow && flow.isMounted()) {
            flow.focus();
        }
    },

    closeSignup() {
        if (document.removeEventListener) {
            document.removeEventListener('keydown', this.closeSignupListener);
        } else {
            window.detachEvent('onkeydown', this.closeSignupListener);
        }
        LoginSignup.closeSignup(this.world());
    },

    closeSignupListener(event) {
        if (event.keyCode === 27) {
            this.closeSignup();
        }
    },

    onResize: function(dims) {
        Client.setDimensions(this.world(), dims);
        setTimeout(this.updateBodyClasses, 1);
    },

    updateBodyClasses() {
        var classes = {
            mobile: Client.isMobile(this.world()),
            'nav-menu': NavMenu.isMobileNavMenuOpen(this.world()),
            // TODO: this may need to be more general in the future
            'modal-open': LoginSignup.isModalSignupOpen(this.world()),
            ...Client.browserClasses()
        };
        var bodyClasses = _.extend(classes, this.pageWrapperFeatures());
        _.each(bodyClasses, (value, key) => {
            if (value) {
                this.addClass(document.body, key);
            } else {
                this.removeClass(document.body, key);
            }
        });
    },

    addClass(el, className) {
        if (el.classList) {
            el.classList.add(className);
        } else {
            if (new RegExp('(^|\\s)' + className + '(\\s|$)', 'gi').test(el.className) === false) {
                el.className += ' ' + className;
            }
        }
    },

    removeClass(el, className) {
        if (el.classList) {
            el.classList.remove(className);
        } else {
            el.className = el.className.replace(new RegExp('(^|\\s)' +
                                                            className +
                                                            '(\\s|$)', 'gi'), ' ');
        }
    },
    pageWrapperFeatures() {
        var pageSpecificFeatures = this.pageComponentType().pageWrapperFeatures;
        return _.extend({}, this.defaultPageWrapperFeatures, pageSpecificFeatures);
    },

    defaultPageWrapperFeatures: {
        'has-mobile-subheader': true,
        'has-sub-header': true,
        'has-search': true
    },

    pageComponent: function () {
        return React.createElement(this.pageComponentType(), {});
    },

    pageComponentType: function () {
        var route = Router.routeForWorld(this.world());
        return Router.componentTypeForRoute(route);
    }
});

module.exports = Layout;
