/* eslint-disable no-console */
const Immutable = require('immutable');
const _ = require('underscore');

const config = require('config');
const Logger = {};

const preBoundFunctions = {};

// Create togglable proxies for all console methods to prevent logging in prod.
var useLogs = false;



[
    'log',
    'warn',
    'debug',
    'info',
    'error',
    'count',
    'timeEnd',
    'time'
].forEach(key => {
    // TODO: in case of missing console functionality, should we polyfill?
    preBoundFunctions[key] = (global.console && global.console[key])
        ? Function.prototype.bind.call(global.console[key], global.console)
        : () => {};

    Logger[key] = (...args) => {
        if (global.console) {
            // Logs are always on for devmode and server-side, togglable
            // otherwise.
            if (__DEV__ || global.__server || useLogs) {
                preBoundFunctions[key](...args);
            }
        }
    };
});

// Expose the ability to toggle logs to global for devs to use.
global.toggleLogs = () => {
    useLogs = !useLogs;
};

// Utility to add tags to the log
function extendTags(context, loggerMethod) {
    return function(value, options, callback) {
        if (typeof options === 'object') {
            var {extra, ...otherOptions} = options;
        } else {
            var extra = {};
            var otherOptions = {};
        }

        // Call the proper raven method with added tags
        context[loggerMethod].call(
            context,
            value,
            Object.assign(
                {
                    tags: { app: 'xrsp' },
                    extra
                },
                otherOptions
            ),
            callback
        );
    };
}

Logger.setupServerErrorLogger = function(serverConfig, raven=undefined) {
    var errorLogger;

    if (!serverConfig.logging['sentry.url']) {
        // Use std logs if sentry isn't configured
        var logFunction = global.console.error.bind(global.console);
        errorLogger = {
            captureError: logFunction,
            captureException: logFunction,
            captureMessage: logFunction
        };

    } else {
        // Proxy logger to raven if it's configured
        var ravenClient = new raven.Client(serverConfig.logging['sentry.url'], {
            logger: 'xrsp-server'
        });

        errorLogger = {
            captureError: extendTags(ravenClient, 'captureError'),
            captureMessage: extendTags(ravenClient, 'captureMessage')
        };
        ravenClient.patchGlobal(function() {
            global.console.log('* Exiting due to uncaught exception');
            process.exit(1);
        });
    }

    Logger.errorLogger = errorLogger;
};

Logger._extra = Immutable.Map({});

Logger.addExtra = function(path, extra) {
    Logger._extra = Logger._extra.setIn(path, Immutable.fromJS(extra));
};

Logger.updateExtra = function(path, callback) {
    Logger._extra = Logger._extra.updateIn(path, callback);
};

Logger.setupClientErrorLogger = function(world) {
    if (!Raven) return;

    const Errors = require('dealers/Errors');

    Raven.config(
        global.location.protocol + '//public@' + global.location.host + '/sentry/42',
        {
            logger: 'xrsp-client',
            tags: { app: 'xrsp' },
            dataCallback: function(data) {
                data.extra = Logger._extra.merge(data.extra).toJS()
                return data;
            }
        }
    ).install();

    // Proxy onerror and get more data about malformed throws
    var windowOnError = window.onerror;
    window.onerror = _.before(5, function() {
        for (var i = 0; i < arguments.length; i++) {
            if (arguments[i] instanceof Error) {
                return windowOnError.apply(window, arguments);
            }
        }

        // Handle malformed throws (possibly from external sources)
        Logger.errorLogger.captureMessage('Malformed Error', {
            extra: {
                onerror: Array.prototype.slice.call(arguments, 0)
            }
        });
    });

    Logger.errorLogger = {
        captureError: _.before(5, extendTags(Raven, 'captureError')),
        captureMessage: _.before(5, extendTags(Raven, 'captureMessage')),
        captureException: _.before(5, extendTags(Raven, 'captureException'))
    };
    global.document.addEventListener('ravenSuccess', () => {
        Errors.SetLastSentryId(world);
    });

    return Logger.errorLogger;
};

// Express app logging middleware
Logger.sentryLogger = function(err, req, res, next) {
    var ws = Immutable.Map({
        Request: {
            originalUrl: req.originalUrl,
            headers: req.headers,
            cookies: req.cookies
        }
    });
    if (req.world) {
        ws = ws.merge(req.world.worldState)
            .deleteIn(['ServerOnly', 'Response']);
    }
    Logger.errorLogger.captureError(err, {
        extra: {
            world: ws.toJS()
        },
        tags: {
            app: 'xrsp'
        }
    });
    res.status(500).send('<h1>' + err.message + '</h1><pre>' + err.stack + '</pre>');
};

// Stathat reporting
const superagent = require('superagent');

function statHandler(err, res) {
    if (err) {
        Logger.errorLogger.captureException(err);
        Logger.error(err);
    } else if (!res.ok || res.body.status !== 200) {
        Logger.error("Stat error:", res.body.msg || res.status);
    } else {
        Logger.log("Stat success");
    }
}

Logger.stats = {
    count(statName, count = 1) {
        Logger.log("stat(count):", statName, count);
        this._request('c', statName, {count});
    },
    value(statName, value) {
        Logger.log("stat(value):", statName, value);
        this._request('v', statName, {value});
    },
    /**
     * Time the execution of fn
     * if callback is passed the fn
     * is assumed to be asynchronous.
     * Callback must be bound.
     */
    time(statName, fn, callback) {
        let now = Date.now();
        if (callback) {
            return fn(() => {
                this.value(statName, Date.now() - now);
                callback.apply(null, arguments);
            });
        } else {
            let result = fn();
            this.value(statName, Date.now() - now);
            return result;
        }
    },
    _request(url, statName, value) {
        // setTimeout and avoid blocking UI actions
        setTimeout(function() {
            if (__DEV__) {
                Logger.log("Ignoring stat request", statName, value);
                if (config.logging['stathat.stats.' + statName] === undefined) {
                    throw new Error("Attempt to use undefined stat: " + statName);
                }
                return;
            }
            superagent.get('http://api.stathat.com/' + url)
                .query({
                    key: config.logging['stathat.stats.' + statName],
                    ukey: config.logging['stathat.user']
                })
                .query(value)
                .end(statHandler);
        }, 0);
    }
};

module.exports = Logger;
