"use strict";

var _ = require('underscore');
var queryString = require('query-string');
var Immutable = require('immutable');
var uriTemplate = _.memoize(require('uri-templates'));
var PageComponentLoader = require('PageComponentLoader');

var config = require('config');

var apiRoutingTable = [
    {url:"/_data/org-intake-lookup", endpoint:"_data/org-intake-lookup"},
    {url:"/_data/{lang}/v1/countries", endpoint:"geo/countries"},
    {url:"/_data/{lang}/v1/countries/{country_geoid}", endpoint:"geo/regions"},
    {url:"/_data/{lang}/v1/locations{?type,q}", endpoint:"geo/search"},
    {url:"/api/v2/info/{+path}", endpoint: "api/v2/info"},
    {url:"/api/v2/info/navigation/{+path}", endpoint: "api/v2/info-navigation"},
    {url:"/api/v2/orgs", endpoint:"api/v2/orgs/create"},
    {url:"/api/v2/sessions/create", endpoint:"api/v2/sessions/create"},
    {url:"/api/v2/users", endpoint:"api/v2/users"},
    {url:"/api/v2/users/check{?email}", endpoint:"api/v2/user/check"},
    {url:"/api/v2/users/info", endpoint:"user"},
    {url:"/api/v2/users/menu-assets", endpoint:"user/assets"},
    {url:"/forgot-password", endpoint:"forgot-password"},
    {url:"/org-pricing-summary{?org_type,country_code,state_code,language_code}", endpoint:"org-pricing-summary"},
    {url:"/view/{id}/fair-radness", endpoint:"cool/fair/bro"},
    {url:"/view/{id}/page-actions", endpoint:"api/page-actions"},
    {url:"/_api/v3/stats{?orgcount}", endpoint:"api/v3/stats"},
    {url:"/_api/v3/constants{?restrictiveRTE,areaOfFocus}", endpoint:"api/v3/constants"},
    {url:"/_api/v3/throw-error", endpoint:"throw-error"}
];

var navigationRoutingTable = [
    {url:"/#", endpoint:"dummy"},
    {url:"/about-us", endpoint:"about-us"},
    {url:"/act", endpoint:"act"},
    {url:"/add-org", endpoint:"add-org"},
    {url:"/contact-us{?sentryId}", endpoint:"contact-us"},
    {url:"/create-content", endpoint:"content/create"},
    {url:"/donate", endpoint:"donate"},
    {url:"/engage/token", endpoint:"engage-token"},
    {url:"/explore/{id}/exhibitors", endpoint:"fair/exhibitors"},
    {url:"/info/{+path}", endpoint:"info/view-path"},
    {url:"/logout", endpoint:"logout"},
    {url:"/me", endpoint:"me/view"},
    {url:"/me/dashboard", endpoint:"me/dashboard"},
    {url:"/me/dashboard/#connections", endpoint:"me/dashboard/connections"},
    {url:"/me/dashboard/#groups", endpoint:"me/dashboard/groups"},
    {url:"/me/messages", endpoint:"me/messages"},
    {url:"/our-team", endpoint:"our-team"},
    {url:"/playground", endpoint:"playground"},
    {url:"/resend-verification{?login}", endpoint:"resend-verification"},
    {url:"/search/v2{?search_type,csrc}", endpoint:"search"},
    {url:"/simulate-error", endpoint:"simulate-error"},
    {url:"/step/{step}/{segment}/{identifier}", endpoint:"step/track"},
    {url:"/step/{step}/{segment}/{identifier}/{+path}", endpoint:"step/track/passthrough"},
    {url:"/view/fair/{id}", endpoint:"view/fair"},
    {url:"/view/fair/{id}/exhibitor", endpoint:"view/fair/exhibitor"},
    {url:"/view/org/{id}", endpoint:"view/org"},
    {url:"/working-here", endpoint:"working-here"}
];

var routingTable = _.memoize((function() {
    if (global.__server) {
        apiRoutingTable = apiRoutingTable.map((route) => {
            return {
                url: internalUrlPrefix() + route.url,
                endpoint: route.endpoint
            };
        });
    }
    return apiRoutingTable.concat(navigationRoutingTable);
}));

var urlPatternsByEndpoint = _.memoize(function() {
    return _.object(routingTable().map(
        (entry) => [ entry.endpoint, entry.url ]
    ));
});

var parametersByMatchingPatternWithUrl = _.memoize(
    (pattern, url) => uriTemplate(pattern).fromUri(url),
    (pattern, url) => [pattern, url]
);

function externalUrlPrefix () {
    return config.idealist['idealist.external_url'];
}

function internalUrlPrefix () {
    return config.idealist['idealist.api.url'];
}

var Router = {
    componentTypeForRoute: function componentTypeForRoute(route) {
        return PageComponentLoader.loadByEndpoint(route.get('endpoint'));
    },

    urlForRoute: function(route, options) {
        options = _.defaults(options || {}, {
            external: false
        });

        var urlPattern = urlPatternsByEndpoint()[route.get('endpoint')];
        var url = uriTemplate(urlPattern).fillFromObject(route.get('parameters').toJS());

        if (options.external) {
            return externalUrlPrefix() + url;
        } else {
            return url;
        }
    },

    routeForUrlPath: function(urlPath, params) {
        urlPath = urlPath.replace(/\/$/, '');
        var entry0 = _.find(routingTable(), (entry) => parametersByMatchingPatternWithUrl(entry.url, urlPath));
        if (entry0) {
            let route = this.route(entry0.endpoint, parametersByMatchingPatternWithUrl(entry0.url, urlPath));
            // fill additional parameters
            return route.set('additionalParameters', Immutable.fromJS(params));
        }
    },

    route: function(endpoint, parameters) {
        parameters = parameters || {};
        return Immutable.fromJS({
            endpoint: endpoint,
            parameters: parameters
        });
    },

    routeForWorld: function(world) {
        return world.get('Route');
    },

    routeIsSuitableTargetForClientNavigation: function (route) {
        return this.componentTypeForRoute(route) !== undefined;
    }
};


module.exports = Router;
