/*
  Mixin: a structure for managing a stack of sequence states, such that
  you can page backwards through states if mutations allow for it.

  Required props on component to be mixed in with:
    - value: the thing that is mutated throughout the undo stack
    - onChange: function to push new state back through after mutation

  Used by ABC.MutableSequence.
*/
module.exports = {
    canUndo: function () {
        return _.last(this.undoStates()).canUndo;
    },

    undo: function (evt) {
        if (evt) { evt.preventDefault(); }

        if (this.canUndo()) {
            this.changeToValue(_.initial(this.undoStates()));
        }
    },

    mutate: function (canUndo, mutator) {
        // Note that this method in only works with values that are
        // mutable in place, such as Arrays, as it does not support
        // reassignment of the newState.value. This could be fixed
        // for Number or String values by making an analogous 'replace'
        // method that assigns the mutator's return to newState.value.
        var newState = {
            value: _.clone(_.last(this.undoStates()).value),
            canUndo: canUndo
        };
        mutator(newState.value);
        this.changeToValue(this.undoStates().concat([newState]));
    },

    changeToValue: function (newStateArray) {
        // FIXME:dh Tacking wireRepresentation onto this Array is a hack
        // to render the value from the undo state before posting to the
        // server. When iterating on this mixin, we should come up with a
        // cleaner way to do this, possibly making a Stack implementation
        // that exposes a 'valueOf' method.
        newStateArray.wireRepresentation = function () {
            if(ABC.valueIsUndoStack(this)) {
                return _.last(this).value;
            } else {
                return this;
            }
        }.bind(newStateArray);
        this.props.value = newStateArray;
        this.props.onChange(newStateArray);
    },

    currentValue: function() {
        return _.last(this.undoStates()).value;
    },

    undoStates: function () {
        var v = this.props.value;
        if(ABC.valueIsUndoStack(v)) {
            return v;
        } else {
            this.props.value = [{value:v, canUndo:false}];
            return this.props.value;
        }
    }
};
