'use strict';

var _ = require('underscore');
var superagent = require('superagent');
var Immutable = require('immutable');
var React = require('react');

var {I18N} = require('I18N');
var FormController = require('FormController');
var Logger = require('Logger');
var Router = require('Router');
var API = require('API');

var Analytics = require('dealers/AnalyticsDealer');

var Org = require('schemas/Org');


class OrgIntakeForm extends FormController {

    pricingMarkup(world) {
        var orgType = this.getField('type');
        var address = this.getAddress();
        var countryCode = address.get('countryCode') || 'US';
        var stateCode = this.stateCode(address);
        var locale = world.get('Locale');

        return API.legacyGet(world,
            ['OrgIntake', 'Pricing', countryCode, orgType, locale, stateCode],
            Router.route('org-pricing-summary',
                {country_code: countryCode, org_type: Org.CAPITALIZED_TYPES[orgType], language_code: locale, state_code: stateCode}));
    }

    resetState() {
        this.state.setDefault();
    }

    // initForm and `get state()` are the only two places that
    // should directly refer to `this._state`
    // all other uses should use the getter.
    initForm(rootComponent) {
        this._state = Org.fromDefaults();
        this.root = rootComponent;
    }

    get state() {
        return this._state;
    }

    getMember(path) {
        let root = this.state;
        if (typeof path === 'string') {
            return root.members[path];
        } else {
            return path.reduce((el, p) => el.members[p], root);
        }
    }

    _set(path, val) {
        return this.getMember(path).set(val);
    }

    setField(fieldOrPath, val) {
        this._set(fieldOrPath, val);
        this.forceUpdateRoot();
    }

    setMany(pairs) {
        let success = pairs.reduce((success, value, key) => {
            let result = this._set(key, value);
            return success && result;
        }, true);
        this.forceUpdateRoot();
        return success;
    }

    getField(fieldOrPath) {
        let member = this.getMember(fieldOrPath);
        if (!member) {
            Logger.warn('getField found no member ' + fieldOrPath);
        } else {
            return member.value;
        }
    }

    getStateAsPayload() {
        var type = this.getField('type');
        if (!type) {
            throw new Error('JSON-API spec requires a type.');
        }
        return {
            data: {
                type: type,
                attributes: this.state.value
            }
        };
    }

    emptyDeduperResults(world) {
        this._setDedupeResults(world, Immutable.List());
    }

    getAddress() {
        var address = this.getField(['addresses', 0]);
        return address || Immutable.Map({});
    }

    // TODO: CAB: maybe a mixin with more functionality.
    haveLocalStorage() {
        return typeof window !== 'undefined' && !!window.localStorage;
    }

    validateOrgHeader() {
        return (
            this.state.members.type.validate() &&
            this.state.members.addresses.validate()
        );
    }

    loadFromLocalStorage() {
        try {
            this._lockLocalStorage = true;
            var cached = this.haveLocalStorage() && localStorage.getItem('OrgIntake');
            if (cached) {
                // let the king's taster check localstorage,
                // if it is not poisoned, the king shall eat.
                let tester = new Org();
                let deserialized = JSON.parse(cached);
                tester.set(deserialized);
                if (tester.members.addresses.validate() && tester.members.type.validate()) {
                    // We have watchers on the type field and the [addresses, 0] field,
                    // so explicitly call setField on those 2 so that changes from localStorage
                    // are propagated to the view layer.
                    // XRSP-236: fix this in hosted conformist by working with react for data flow.
                    this.setField('type', deserialized.type);
                    this.setField(['addresses', 0], deserialized.addresses[0]);
                } else {
                    this.updateLocalStorage();
                }
            } else {
                this.updateLocalStorage();
            }
        } catch (e) {
            // malformed JSON, ignore.
            Logger.warn(e);
        } finally {
            this._lockLocalStorage = false;
        }
    }

    updateLocalStorage() {
        if (!this._lockLocalStorage && this.haveLocalStorage()) {
            try {
                localStorage.setItem(
                    'OrgIntake', JSON.stringify({
                        type: this.state.members.type.value,
                        addresses: this.state.members.addresses.value
                    })
                );
            } catch (e) {}
        }
    }

    stateCode(address) {
        if (address.get('countryCode') !== 'US') {
            return '';
        }

        if (address.get('fallback')) {
            return address.get('regionCode') ? address.get('regionCode').split('.')[1] : '';
        } else {
            let locationParts = address.get('formatted').split(', ');
            if (locationParts.length < 3) {
                throw new Error('What kind of address is ' + address.get('formatted'));
            }
            return locationParts[locationParts.length - 2]; // -1
        }
    }

    dedupeResultsFor(world, term) {
        if (term.length <= 2) {
            this.emptyDeduperResults(world);
            return;
        }

        var url = Router.urlForRoute(Router.route('_data/org-intake-lookup'));
        var location = this.getAddress().get('location');
        if (!location) {
            location = this.getAddress().get('regionName')
                + ', ' + this.getAddress().get('countryName');
        }
        superagent.get(url).timeout(10000)
            .query({
                q: term,
                geoid: this.getAddress().get('geoid'),
                location_string: location,
                type: this.getField('type'),
            })
            .end(function (res) {
                var dedupes = Immutable.List();
                if (res.status === 200) {
                    dedupes = Immutable.fromJS(res.body);
                }
                this._setDedupeResults(world, dedupes);
            }.bind(this));
    }

    createNewOrgEntryForTerm(term) {
        return {
            isAddNew: true,
            name: term
        };
    }

    _setDedupeResults(world, results) {
        world.performAction({
            name: 'OrgIntakeForm/dedupeResults',
            transaction: (ws) =>
                ws.setIn(['OrgIntake','dedupeResults'], results)
        });
    }

    getDedupeResults(world, term) {
        let results = world.getIn(['OrgIntake', 'dedupeResults']);

        if (results && !results.find(res => res.get('isExact'))) {
            results = results.unshift(Immutable.fromJS(this.createNewOrgEntryForTerm(term)));
        }

        return results;
    }

    selectOrg(orgObject) {
        this._selectedOrg = orgObject;
        this.forceUpdateRoot();
    }

    selectCurrentName() {
        this.selectOrg(this.createNewOrgEntryForTerm(this.state.members.fullName.value));
    }

    selectedOrgItem() {
        return this._selectedOrg;
    }

    existingOrgIsSelected() {
        return typeof this._selectedOrg !== 'undefined' && !!this._selectedOrg.url;
    }

    existingOrgWasReset(world) {
        this.emptyDeduperResults(world);
        this._selectedOrg = undefined;
        this._orgIsResetting = true;
        this.forceUpdateRoot();
    }

    v2WireRepresentation() {
        let payload = Immutable.fromJS(this.state.value);
        // First the fields that haven't changed name in v3.
        var rep = _.pick(payload.toJS(), 'ein', 'keywords', 'website');
        var addr = this.getAddress().toJS();

        // amazing requirement on the server that we have underscores in these
        if (_.isString(addr.regionCode)) {
            addr.regionCode = addr.regionCode.replace('.','_');
        }

        return Object.assign(rep, {
            'org_type': Org.CAPITALIZED_TYPES[payload.get('type')],
            'short_name': payload.get('shortName'),
            'name': payload.get('fullName'),
            'mission': payload.get('description'),

            'address_geo_autocomplete_location': addr.location,
            'address_geo_autocomplete_geoid': addr.geoid,
            'address_geo_mode': addr.fallback ? 'fallback' : 'autocomplete',
            'address_geo_fallback_city': addr.city,
            'address_geo_fallback_state': addr.regionCode,
            'address_geo_fallback_country': addr.countryCode,

            'address_country_name': addr.countryName,
            'address_state_name': addr.regionName,

            'address_street_address': payload.get('streetAddress'),
            'address_delivery_details': payload.get('deliveryDetails'),

            'logo': payload.get('imageUrl'),
            'activation_actions': payload.get('activationActions')
        });
    }

    v3SubmitOrg(world) {
        // Each org type has a different route.
        var orgType = this.state.get('type');
        var route = Router.route('api/v3/' + orgType + 's/create');
        var org = this.getStateAsPayload();
        this._submitOrg(world, route, org);
    }

    v2SubmitOrg(world) {
        var route = Router.route('api/v2/orgs/create');
        var org = this.v2WireRepresentation();
        this._submitOrg(world, route, org);
    }

    get submittingPath() {
        return ['OrgIntake', 'submitting'];
    }

    isSubmitting(world) {
        return world.getIn(this.submittingPath);
    }

    _submitOrg(world, route, org) {
        if (!this.state.validate()) {
            this.state.addError(Org.SERVER_ERROR('VALIDATION'));
            return;
        }
        // wtf?
        this.setMany(Immutable.fromJS(this.state.value));
        var self = this;
        Logger.stats.count('org.submission.attempt');
        world.performAction({
            name: 'SubmitOrgIntake',
            transaction: ws => ws.setIn(this.submittingPath, true)
        });
        API.request(world, route,
            (data, headers) => {return {
                name: 'submitOrg/success',
                transaction: ws => ws,  // NOTE:eo don't set submitting to False
                                        // because the page won't redirect immediately
                                        // so keep the spinner going until they bounce.
                effect: () => {
                    Logger.stats.count('org.submission.success');
                    Analytics.trackPageview('/virtual/add-org/complete', () => {
                        document.location.href = headers.get('location');
                    });
                }
            };},
            (error) => {return {
                name: 'submitOrg/error',
                transaction: (ws) => ws.setIn(this.submittingPath, false),
                effect: () => {
                    Logger.stats.count('org.submission.failure');
                    // TODO: this will change when we switch to api/v3
                    if (error.get('status') !== 400) {
                        // unauthorized, or some other problem with user
                        this.state.valid = false;
                        this.state.addError(Org.SERVER_ERROR('ACCOUNT'));
                    } else {
                        // validation error
                        let fieldErrors = error.getIn(['body', 'fieldErrors']);
                        let fieldMembers = {
                            'name': self.getMember('fullName'),
                            'address_geo_fallback_city': self.getMember(['addresses', 0, 'city'])
                        };
                        let mbr;
                        fieldErrors && fieldErrors.forEach((errors, field) => {
                            mbr = fieldMembers[field] || this.state.members[field];
                            errors.forEach(e => mbr.addError(e));
                        });
                        // monkeypatch the form's valid prop because running validate would
                        // clobber our new errors
                        if (fieldErrors.size) {
                            this.state.addError(Org.SERVER_ERROR('VALIDATION'));
                            this.state.valid = false;
                        }
                    }
                }
            };},
            {method: 'POST', payload: org}
        );
    }
}

global.__form__ = module.exports = new OrgIntakeForm();
