var React = require('react');
var Input = require('elements/Forms/Input');

var guid = 0;
var k = function(){};
var eq = function(a,b){return a===b;};
var addClass = require('./add-class');
var ComboboxOption = require('./option');

function getLabel(component) {
    var hasLabel = component.props.label != null;
    return hasLabel ? component.props.label : component.props.children;
}

function matchFragment(userInput, firstChildLabel) {
    userInput = userInput.toLowerCase();
    firstChildLabel = firstChildLabel.toLowerCase();
    if (userInput === '' || userInput === firstChildLabel) {
        return false;
    }
    if (firstChildLabel.toLowerCase().indexOf(userInput.toLowerCase()) === -1) {
        return false;
    }
    return true;
}


module.exports = React.createClass({
    displayName: 'Combobox',

    propTypes: {

        /**
         * Defaults to 'both'. 'inline' will autocomplete the first matched Option
         * into the input value, 'list' will display a list of choices, and of
         * course, both does both (do you have a soft 'L' in there when you say
         * "both" out loud?)
        */
        autocomplete: React.PropTypes.oneOf(['both', 'inline', 'list']),

        /**
         * Called when the combobox receives user input, this is your chance to
         * filter the data and rerender the options.
         *
         * Signature:
         *
         * ```js
         * function(userInput){}
         * ```
        */
        onInput: React.PropTypes.func,

        /**
         * Called when the combobox receives a keyUp of user input, as opposed to
         * onInput which goes off on change events only. This callback gets passed
         * the raw event.
         *
         * Signature:
         *
         * ```js
         * function(event){}
         * ```
        */
        onInputKeyUp: React.PropTypes.func,


        /**
         * Called when the combobox receives a selection. You probably want to reset
         * the options to the full list at this point.
         *
         * Signature:
         *
         * ```js
         * function(selectedValue){}
         * ```
        */
        onFocusOption: React.PropTypes.func,

        /**
         * Called when the combobox is blurred.
         *
         * Signature:
         *
         * ```js
         * function(){}
         * ```
        */
        onBlur: React.PropTypes.func,

        /**
         * Called when the combobox is *closed* via making a selection (with
         * the enter key or a click).
         *
         * Signature:
         *
         * ```js
         * function(selectedValue){}
         * ```
        */
        onSelect: React.PropTypes.func,

        /**
         * Function used to compare Option values instead of the default `===`.
         *
         * Signature:
         *
         * ```js
         * function(value1, value2){} => bool
         * ```
        */
        valueComparator: React.PropTypes.func,

        /**
         * The initial value of the component.
        */
        value: React.PropTypes.string,

        /**
         * The tabIndex of the autocomplete input.
        */
        tabIndex: React.PropTypes.string,

        /**
         * The max-height to apply to the menu of autocomplete options
        */
        maxMenuHeight: React.PropTypes.number
    },

    getDefaultProps: function() {
        return {
            autocomplete: 'both',
            onInput: k,
            onFocusOption: k,
            onBlur: k,
            onSelect: k,
            valueComparator: eq,
            value: null,
            maxMenuHeight: Infinity
        };
    },

    getInitialState: function() {
        return {
            value: this.props.value,
            // the value displayed in the input
            inputValue: this.findInputValue(),
            isOpen: false,
            focusedIndex: null,
            matchedAutocompleteOption: null,
            // this prevents crazy jumpiness since we focus options on mouseenter
            usingKeyboard: false,
            activedescendant: null,
            menu: {
                children: [],
                activedescendant: null,
                isEmpty: true
            }
        };
    },

    componentWillMount: function() {
        this.setState({menu: this.makeMenu()});
    },

    componentWillReceiveProps: function(newProps) {
        var newState = {
            menu: this.makeMenu(newProps.children)
        };
        if (newProps.value !== null && newProps.value !== undefined) {
            newState.inputValue = this.findInputValue(newProps.value);
        }
        this.setState(newState);
    },

    // ***** PART THE FIRST: turn browser events into 2nd-order events ********

    handleButtonClick: function() {
        this.state.isOpen ? this.hideList() : this.showList();
        this.focusInput();
    },

    handleKeydown: function(event) {
        var handlerName = this.inputKeydownMap[event.keyCode];
        if (!handlerName) {
            return;
        }
        if (this.state.isOpen) {
            event.preventDefault();
        } else if (event.keyCode === 9) {
            return;
        }
        this.setState({usingKeyboard: true});
        this[handlerName].call(this, event);
    },

    handleOptionKeyDown: function(child, event) {
        var handlerName = this.optionKeydownMap[event.keyCode];
        // if (this.state.isOpen && event.keyCode === 9) {
        //     return;
        // }
        if (!handlerName) {
            // if the user starts typing again while focused on an option, move focus
            // to the input, select so it wipes out any existing value
            if (event.keyCode !== 9) {
                this.selectInput();
            } else {
                // unless they hit tab, because tab doesn't do stuff in selects
                event.preventDefault();
            }
            return;
        }
        event.preventDefault();
        this.setState({usingKeyboard: true});
        this[handlerName].call(this, child);
    },

    handleOptionMouseEnter: function(index) {
        if (this.state.usingKeyboard) {
            this.setState({usingKeyboard: false});
        } else {
            this.focusOptionAtIndex(index);
        }
    },

    handleInputKeyUp: function(event) {
        if (this.props.onInputKeyUp && this.props.onInputKeyUp(event)===false) {
            return;
        }
        if (
            this.state.menu.isEmpty ||
            // autocompleting while backspacing feels super weird, so let's not
            event.keyCode === 8 /*backspace*/ ||
            !this.props.autocomplete.match(/both|inline/)
        ) {
            return;
        }
        this.autocompleteInputValue();
    },

    handleInputChange: function() {
        var value = this._input.value;
        this.clearSelectedState(function() {
            this.props.onInput(value);
            if (!this.state.isOpen) {
                this.showList();
            }
        }.bind(this), {inputValue: value});
    },

    handleInputBlur: function(event) {
        var focusedAnOption = this.state.focusedIndex != null;
        if (focusedAnOption) {
            return;
        }
        this.maybeSelectAutocompletedOption();
        this.hideList(event);
    },

    inputKeydownMap: {
        9: 'selectFirstOption',
        13: 'selectOnEnter',
        27: 'hideOnEscape',
        38: 'focusPrevious',
        40: 'focusNext'
    },

    optionKeydownMap: {
        13: 'selectOptionAndClose',
        27: 'hideOnEscape',
        38: 'focusPrevious',
        40: 'focusNext'
    },

    // **** PART THE SECOND: dispatch on second-order events *********

    handleOptionBlur: function(event) {
        // don't want to hide the list if we focused another option
        this.blurTimer = setTimeout(function() {
            this.hideList(event);
        }.bind(this), 0);
    },

    handleOptionFocus: function() {
        // see `handleOptionBlur`
        clearTimeout(this.blurTimer);
    },

    /**
     * Events that require the dropdown menu to check its size. For example, if
     * the dropdown menu appears at the bottom of the page, it may be very
     * short. See the showList() and hideList() methods.
     */
    menuEvents: [
        'resize',
        'scroll',
        'orientationchange'
    ],

    showList: function() {
        if (this.props.autocomplete.match(/both|list/)) {
            this.setState({isOpen: true});
        }
        this.addMenuEventHandlers();
    },

    hideList: function(event) {
        if (this.isMounted()) {
            this.setState({isOpen: false});
        }
        this.removeMenuEventHandlers();
        this.props.onBlur(event);
    },

    addMenuEventHandlers: function() {
        this.menuEvents.forEach(eventName => {
            if (window.addEventListener) {
                window.addEventListener(eventName, this.handleMenuEvent);
            } else {
                window.attachEvent('on' + eventName, this.handleMenuEvent);
            }
        });
    },

    removeMenuEventHandlers: function() {
        this.menuEvents.forEach(eventName => {
            if (window.removeEventListener) {
                window.removeEventListener(eventName, this.handleMenuEvent);
            } else {
                window.detachEvent('on' + eventName, this.handleMenuEvent);
            }
        });
    },

    handleMenuEvent: function(evt) {
        this.forceUpdate();
    },

    componentWillUnmount: function() {
        this.removeMenuEventHandlers();
    },

    hideOnEscape: function(event) {
        this.hideList(event);
        this.focusInput();
    },

    focusInput: function() {
        this._input.focus();
    },

    selectInput: function() {
        this._input.select();
    },

    selectOnEnter: function() {
        this.maybeSelectAutocompletedOption();
        this.selectInput();
    },

    maybeSelectAutocompletedOption: function() {
        if (!this.state.matchedAutocompleteOption) {
            return;
        }
        this.selectOption(this.state.matchedAutocompleteOption, {focus: false});
    },

    /**
    * We don't create the <ComboboxOption> components, the user supplies them,
    * so before rendering we attach handlers to facilitate communication from
    * the ComboboxOption to the Combobox.
    */
    makeMenu: function(children) {
        var activedescendant;
        var isEmpty = true;
        children = children || this.props.children;
        var newChildren = React.Children.map(children, function(child, index) {
            if (child === null || child.type !== ComboboxOption) {
                // allow random elements to live in this list
                return child;
            }
            isEmpty = false;

            var valueMatch = this.props.valueComparator(
                this.state.value, child.props.value
            );
            // need an ID for WAI-ARIA
            var newId = valueMatch ?
                (child.props.id || 'rf-combobox-selected-'+(++guid))
                :
                child.props.id;
            if (valueMatch) {
                activedescendant = newId;
            }
            return React.cloneElement(child, {
                id: newId,
                isSelected: valueMatch,
                onBlur: function(evt) {this.handleOptionBlur(evt);}.bind(this),
                onClick: this.selectOption.bind(this, child, {hide: true}),
                onTouchEnd: function() {
                    if (!this.cancelSelect) { this.selectOption(child, {hide: true}); }
                }.bind(this),
                onFocus: this.handleOptionFocus,
                onKeyDown: evt => this.handleOptionKeyDown(child, evt),
                onMouseEnter: this.handleOptionMouseEnter.bind(this, index)
            }, child.props.children);
        }.bind(this));
        return {
            children: newChildren,
            activedescendant: activedescendant,
            isEmpty: isEmpty
        };
    },

    getClassName: function() {
        var className = addClass(this.props.className, 'rf-combobox');
        if (this.state.isOpen) {
            className = addClass(className, 'rf-combobox-is-open');
        }
        return className;
    },

    /**
    * When the user begins typing again we need to clear out any state that has
    * to do with an existing or potential selection.
    */
    clearSelectedState: function(cb, overrides) {
        var newState = Object.assign({}, {
            focusedIndex: null,
            inputValue: null,
            value: null,
            matchedAutocompleteOption: null,
            activedescendant: null
        }, overrides);
        this.setState(newState, cb);
    },

    /**
    * Autocompletes the input value with a matching label of the first
    * ComboboxOption in the list and selects only the fragment that has
    * been added, allowing the user to keep typing naturally.
    */
    autocompleteInputValue: function() {
        if (
            this.props.autocomplete == false ||
            this.props.children.length === 0
        ) {
            return;
        }
        var input = this._input;
        var inputValue = input.value;
        var firstChild = this.props.children.length ?
            this.props.children[0] :
            this.props.children;
        var label = getLabel(firstChild);
        var fragment = matchFragment(inputValue, label);
        if (!fragment) {
            return;
        }
        input.value = label;
        input.setSelectionRange(inputValue.length, label.length);
        this.setState({matchedAutocompleteOption: firstChild});
    },

    selectOption: function(child, options) {
        options = options || {};
        this.setState({
            value: child.props.value,
            inputValue: getLabel(child),
            matchedAutocompleteOption: null
        }, function() {
            this.props.onFocusOption(child.props.value, child);
            if (options.hide === true) {
                this.props.onSelect(child.props.value, child);
                this.hideList();
            }
            if (options.focus === true) {
                this.selectInput();
            }
        }.bind(this));
    },

    selectOptionAndClose: function(child) {
        this.selectOption(child, { hide: true, focus: true });
    },

    selectFirstOption: function(event) {
        if (React.Children.count(this.props.children) > 0
                && this.state.isOpen
                && !event.nativeEvent.shiftKey) {
            event.preventDefault();
            this.focusOptionAtIndex(0, true);
            setTimeout(function() {
                if (this.isMounted()) {
                    this.focusInput();
                }
            }.bind(this), 1);
        }
     },

    focusNext: function(event) {
        if (this.state.menu.isEmpty) {
            return;
        }
        var index = this.state.focusedIndex == null ?
            0
            :
            this.state.focusedIndex + 1;
        this.focusOptionAtIndex(index);
    },

    focusPrevious: function(event) {
        if (this.state.menu.isEmpty) {
            return;
        }
        var last = this.props.children.length - 1;
        var index = this.state.focusedIndex == null ?
            last
            :
            this.state.focusedIndex - 1;
        this.focusOptionAtIndex(index);
    },

    focusSelectedOption: function() {
        var selectedIndex;
        React.Children.forEach(this.props.children, function(child, index) {
            if (this.props.valueComparator(child.props.value, this.state.value)) {
                selectedIndex = index;
            }
        }.bind(this));
        this.showList();
        this.setState({
            focusedIndex: selectedIndex
        }, this.focusOption);
    },

    findInputValue: function(value) {
        value = value || this.props.value;
        // TODO: might not need this, we should know this in `makeMenu`
        var inputValue;
        React.Children.forEach(this.props.children, function(child) {
            if (this.props.valueComparator(child.props.value, value)) {
                inputValue = getLabel(child);
            }
        }.bind(this));
        return inputValue || value;
    },

    focusOptionAtIndex: function(index, andHide) {
        if (!this.state.isOpen && this.state.value) {
            return this.focusSelectedOption();
        }
        this.showList();
        var length = React.Children.count(this.props.children);
        if (index === -1) {
            index = length - 1;
        } else if (index === length) {
            index = 0;
        }
        this.setState({
            focusedIndex: index
        }, this.focusOption);

        // Select the focused option:
        var focusedChild;
        React.Children.forEach(this.props.children, function(child, childIndex) {
            if (childIndex === index) {
                focusedChild = child;
            }
        });
        this.selectOption(focusedChild, {focus: false, hide: andHide});
    },

    focusOption: function() {
        var index = this.state.focusedIndex || 0;
        var node = React.findDOMNode(this.refs.list).childNodes[index];
        node.focus();
    },

    recordBoundingBox: function() {
        var node = this._input;
        this._rect = node.getBoundingClientRect();
    },

    componentDidMount: function() {
        this.recordBoundingBox();
    },
    componentDidUpdate: function() {
        this.recordBoundingBox();
    },

    portalStyles: function() {
        if (!this._input) {
            return {};
        } else {
            var windowHeight;
            if (window.innerHeight) {
                windowHeight = window.innerHeight;
            } else {
                windowHeight = document.documentElement.clientHeight;
            }
            return {
                position: 'absolute',
                display: this.state.isOpen ? 'block' : 'none',
                overflow: 'auto',
                maxHeight: this._rect
                    ? Math.min(this.props.maxMenuHeight, windowHeight - this._rect.bottom)
                    : 0
            };
        }
    },

    findInputFromRef: function(ref) {
        if (ref === null) {
            this._input = null;
            return;
        }

        var el = React.findDOMNode(ref);
        if (el.tagName === 'INPUT') {
            this._input = el;
        } else {
            this._input = el.querySelector('input');
        }
    },

    render: function() {
        return (
            <div className={this.getClassName()}>
                <Input
                    type="text"
                    ref={this.findInputFromRef}
                    placeholder={this.props.placeholder}
                    disabled={this.props.disabled}
                    className="rf-combobox-input"
                    tabIndex={this.props.tabIndex}
                    defaultValue={this.props.value}
                    value={this.state.inputValue}
                    onChange={this.handleInputChange}
                    onBlur={this.handleInputBlur}
                    onKeyDown={this.handleKeydown}
                    onKeyUp={this.handleInputKeyUp}
                    role="combobox"
                    aria-activedescendant={this.state.menu.activedescendant}
                    aria-autocomplete={this.props.autocomplete}
                    aria-owns="react-autocomplete-results"
                />
                <span
                    aria-hidden="true"
                    className="rf-combobox-button"
                    onClick={this.handleButtonClick}
                >▾</span>
                <div style={{position: 'relative'}}>
                    {React.Children.count(this.props.children) > 0 && <div
                        style={this.portalStyles()}
                        id="react-autocomplete-results"
                        ref="list"
                        className="rf-combobox-list"
                        aria-expanded={this.state.isOpen+''}
                        onTouchStart={(evt) => { this.cancelSelect = false; }}
                        onTouchMove={(evt) => { this.cancelSelect = true; }}
                        role="listbox"
                        >{this.state.menu.children}
                    </div>}
                </div>
            </div>
        );
    }
});
